/*
 *
 *  Near Field Communication nfctool
 *
 *  Copyright (C) 2011-2013  Intel Corporation
 *  Copyright (C) 2004-2010  Marcel Holtmann <marcel@holtmann.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <stdbool.h>

bool use_color(void);

#define COLOR_OFF	"\x1B[0m"
#define COLOR_BLACK	"\x1B[0;30m"
#define COLOR_RED	"\x1B[0;31m"
#define COLOR_GREEN	"\x1B[0;32m"
#define COLOR_YELLOW	"\x1B[0;33m"
#define COLOR_BLUE	"\x1B[0;34m"
#define COLOR_MAGENTA	"\x1B[0;35m"
#define COLOR_CYAN	"\x1B[0;36m"
#define COLOR_WHITE	"\x1B[0;37m"
#define COLOR_WHITE_BG	"\x1B[0;47m"
#define COLOR_HIGHLIGHT	"\x1B[1;39m"

#define COLOR_ERROR	"\x1B[1;31m"

#define print_indent(indent, color, fmt, args...) \
do { \
	printf("%*c%s" fmt " %s\n", (indent), ' ', \
		use_color() ? (color) : "", ## args, \
		use_color() ? COLOR_OFF : ""); \
} while (0)

#define print_indent_prefix(indent, color1, prefix, color2, fmt, args...) \
do { \
	printf("%*c%s%s%s" fmt "%s\n", (indent), ' ', \
		use_color() ? (color1) : "", prefix, \
		use_color() ? (color2) : "", ## args, \
		use_color() ? COLOR_OFF : ""); \
} while (0)

int num_columns(void);
